"""
Contains miscellaneous utility functions.
"""

import math
from bisect import bisect_left


def round_down(val, base):
    """
    Round val down to multiple of base.
    """
    return base * math.floor(val / base)


def round_up(val, base):
    """
    Round val up to multiple of base.
    """
    return base * math.ceil(val / base)


def round_nearest(val, base):
    """
    Round val to nearest multiple of base.
    """
    return base * round(val / base)


def take_closest(myList, myNumber, cache={}, granularity=1.0):
    """
    Assumes myList is sorted. Returns closest value to myNumber.
    If two numbers are equally close, return the smallest number.
    Additionally, use cache to conserve runtime.
    Use granularity to round myNumber to cache key.
    Source: https://stackoverflow.com/questions/12141150/from-list-of-integers-get-number-closest-to-a-given-value/12141511#12141511
    """
    myNumberGran = round_nearest(myNumber, granularity)

    if myNumberGran in cache:
        return cache[myNumberGran]

    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return myList[0]
    if pos == len(myList):
        return myList[-1]
    before = myList[pos - 1]
    after = myList[pos]
    if after - myNumber < myNumber - before:
        cache[myNumberGran] = after   # add to cache
        return after
    else:
        cache[myNumberGran] = before   # add to cache
        return before


"""
Contains high-level algorithm code.
"""

import copy
import math
import numpy as np
from random import randint
from tqdm import tqdm

from .genetic_op import *
from .utils import take_closest


def snrpga2(D_m, T_m, D_t, D, E, g_start, t2i, times, C, params):
    """
    High-level code for SNRPGA2 genetic algorithm.

    Parameters:
    - D_m: 3D np.ndarray (time-dependent distance matrix)
    - T_m: 3D np.ndarray (time-dependent time travel matrix)
    - D_t: 1D np.ndarray (delivery times for each location)
    - D: 1D np.ndarray (demand for each location)
    - E: 1D np.ndarray (end time for each location)
    - g_start: float (global start time aka departure from depot)
    - t2i: dict (map time to index of D_m and T_m on axis 0)
    - times: list (times of indices of axis 0 of D_m and T_m)
    - C: float (capacity of each vehicle)
    - params: dict (algorithm hyperparameters)

    Returns:
    - sol_res: tuple (detailed info of best chromosome)
    - sol_score: float (best chromosome's fitness score)
    """

    cache = {}

    # Perform assertion checks to ensure feasibility of solution
    assert params["init_method"] in ['random_sample', 'random_seq', 'tsp']
    assert D_m.shape[1] >= 3, "At least three customers required"
    assert D.max() <= C, "Largest demand must not exceed vehicle capacity"
    for k in range(1, D.shape[0]):
        time_to_k = T_m[0, 0, k]
        t_k = take_closest(times,
                           g_start + time_to_k + D_t[k],
                           cache=cache,
                           granularity=params["cache_gran"])
        i_k = t2i[t_k]
        time_from_k = T_m[i_k, k, 0]
        assert time_to_k + D_t[k] + time_from_k <= E[0] - g_start, \
            "Every customer must be roundtrip reachable from depot within " + \
                "global time frame (accounting for both travel time and " + \
                "waiting time according to customer time windows)"


    # Generate random initial population
    n = D_m.shape[1] - 1  # number of customers
    if params["init_method"] == 'random_sample':
        L = [np.random.choice(np.arange(1, n + 1), n, replace=False)
                for _ in range(params["pop_size"])]
    elif params["init_method"] == 'random_seq':
        i = randint(1, n)
        L = [np.hstack([np.arange(i, n + 1), np.arange(1, i)])
                for _ in range(params["pop_size"])]
    else:  # init == 'tsp'
        # TODO - implement
        pass

    # Evaluate fitness function of init pop
    L_res = [assign_routes(ch, C, D_t, D, E, T_m, g_start, times, t2i,
                             cache=cache, granularity=params["cache_gran"])
                             for ch in L]
    L_scores = [eval_fitness(params["obj_func"], routes=res[0],
                                departure_times=res[3], dist_matrix=D_m,
                                depot_arrivals=res[4], times=times, t2i=t2i,
                                cache=cache, granularity=params["cache_gran"],
                                g_start=g_start, w_t=params["w_t"])
                                for res in L_res]

    pbar = tqdm(range(1, params["mng"] + 1), desc='Generation count')
    for _ in pbar:
        # Select subpopulation
        S_ind = subpop_selection(L_scores)

        # Run tournament selection to generate pairs
        pairs = tournament_selection(S_ind,
                    math.ceil(len(S_ind)), 5, p=params["ts_prob"])

        # Create next generation through each candidate pair
        L_copy = copy.deepcopy(L)
        for pair in pairs:
            i1, i2 = pair[0], pair[1]

            # Produce offspring through crossover and mutation
            c1, c2 = L_copy[i1], L_copy[i2]
            d1, d2, modified_c = crossover(c1, c2, x_prob=params["x_prob"])
            d1, d2, modified_m = mutation(d1, d2, m_prob=params["m_prob"])

            if not modified_c and not modified_m:
                continue  # skip if offspring are identical to parents

            d1_res = assign_routes(d1, C, D_t, D, E, T_m, g_start, times,
                            t2i, cache=cache, granularity=params["cache_gran"])
            d2_res = assign_routes(d2, C, D_t, D, E, T_m, g_start, times,
                            t2i, cache=cache, granularity=params["cache_gran"])

            # Evaluate fitness function of offspring
            d1_score, d2_score = [eval_fitness(params["obj_func"],
                                    g_start=times[0], routes=res[0],
                                    departure_times=res[3], dist_matrix=D_m,
                                    depot_arrivals=res[4], times=times,
                                    t2i=t2i, cache=cache, w_t=params["w_t"],
                                    granularity=params["cache_gran"])
                                    for res in (d1_res, d2_res)]

            # Replace in population if fitness score is lower
            replace_in_pop(L, L_scores, L_res, i1, d1, d1_score, d1_res)
            replace_in_pop(L, L_scores, L_res, i2, d2, d2_score, d2_res)

        pbar.set_description("Best score = %.3f" % min(L_scores))


    # Retrieve information for top chromosome in last population
    sol_ind = subpop_selection(L_scores, n=1, mode='top')[0]
    sol_res, sol_score = L_res[sol_ind], L_scores[sol_ind]

    return sol_res, sol_score



"""
Contains helper functions for computation of values for algorithm input.
"""

import math
import numpy as np
from scipy.spatial import distance_matrix

from .utils import round_down, round_up


def compute_window_size(b_d, e_d, num_loc, matr_size=1000000):
    """
    Compute a value for window_size, that allows adequate granularity
    without necesitating too much memory for the later-to-be-computed
    distance and time matrices.

    Parameters:
    - b_d: float (beginning depot time, >= 0)
    - e_d: float (ending depot time, >= 0)
    - num_loc: int (number of locations)
    - matr_size: int

    Returns:
    - window_size: int
    """
    num_windows = matr_size / (num_loc ** 2)
    window_size = math.ceil((e_d - b_d) / num_windows)
    return window_size


def compute_t2i(b_d, e_d, window_size):
    """
    Construct mapping from raw time to time window index.
    Parameters:
    - b_d: float (beginning depot time, >= 0)
    - e_d: float (ending depot time, >= 0)
    - window_size: int (amount of unit time for each window)
    Returns:
    - t2i: dict
    """
    assert b_d >= 0
    assert e_d >= 0

    t2i = {}
    i = 0
    window_begin = round_down(b_d, window_size)
    window_end = round_up(e_d, window_size)
    while window_begin <= window_end:
        t2i[window_begin] = i
        window_begin += window_size
        i += 1

    return t2i


def compute_i2t(t2i):
    """
    Construct mapping from time window index to raw time.
    Parameters:
    - t2i: dict
    Returns:
    - i2t: list (keys-values from t2i reversed)
    """
    i2t = [0 for _ in range(len(t2i))]
    for k, v in t2i.items():
        i2t[v] = k
    return i2t


def compute_dist_matrix(x, y):
    """
    Compute distance matrix given x and y coords.
    Parameters:
    - x, y: list (representing coordinates)
    Returns:
    - dist_matrix: np.ndarray
    """
    assert len(x) == len(y)
    pts = np.vstack([x, y]).T
    dist_matrix = distance_matrix(pts, pts)
    return dist_matrix


def compute_raw_time_matrix(D_r, scen=None, C=None, S=None, i2t=None):
    """
    Construct raw travel time matrix from distances and speeds
        for each time window. It will be symmetrical.

    Parameters:
    - D_r: 2D np.ndarray  (raw distance matrix)
    - if time-dependent matrix returned, these variables are not None:
        - scen: int  (scenario id, determines degree of time dep)
        - C: 2D np.ndarray  (upper triangular category matrix)
        - S: 3D np.ndarray  (speed matrix, scenario on axis 0)
        - i2t: list

    Returns:
    - T_r: np.ndarray
        - 2D if not time-dependent
        - 3D if time-dependent (departure time window on axis 0) 
    """

    if scen is None:  # non time dependent, all speeds are trivially 1
        return D_r  # 2D
    
    assert scen is not None and C is not None
    assert S is not None and i2t is not None
    assert S.shape[1] == len(i2t)
    assert D_r.shape[0] <= C.shape[0]

    T_r = np.repeat(D_r[np.newaxis, :, :], len(i2t), axis=0)
    
    for p in range(D_r.shape[0] - 1):
        for q in range(p + 1, D_r.shape[1]):
            T_r[:, p, q] *= S[scen, :, C[p, q]]
            T_r[:, q, p] *= S[scen, :, C[p, q]]
            
    return T_r  # 3D


def compute_time_matrix(T, B, i2t):
    """
    Construct travel time matrix that uses (possibly time-dependent)
        travel times, begin and end times for each location, and
        delivery times for each location.

    Parameters:
    - T: np.ndarray (raw travel times between each location)
        - 2D if not time-dependent
        - 3D if time-dependent (departure time window on axis 0)
    - B: 1D np.ndarray (begin times for each location)
    - i2t: list (map time window index to raw time)

    Returns:
    - T_m: 3D np.ndarray
        - T_m[i,s,d] = time to travel from s to d if departure time is i2t[i]
        - may include waiting time
        - does not include delivery time
    """
    
    T_m = np.zeros((len(i2t), T.shape[0], T.shape[1]))

    if T.ndim == 2:  # need to add time-dimension    
        for d in range(T.shape[1]):
            for s in range(T.shape[0]):
                for i, t in enumerate(i2t):
                    if t + T[s, d] < B[d]:  # arrival before start time
                        T_m[i, s, d] = B[d] - t  # waiting time
                    else:  # no need to wait
                        T_m[i, s, d] = T[s, d]
    else:  # T.ndim == 3
        for d in range(T.shape[1]):
            for s in range(T.shape[0]):
                for i, t in enumerate(i2t):
                    if t + T[i, s, d] < B[d]:  # arrival before start time
                        T_m[i, s, d] = B[d] - t  # waiting time
                    else:  # no need to wait
                        T_m[i, s, d] = T[i, s, d]

    return T_m


from setuptools import setup, find_packages

VERSION = '0.3.0' 
DESCRIPTION = 'SNRPGA2 algorithm to solve TDVRPTW'
LONG_DESCRIPTION = 'Implementation of the genetic algorithm SNRPGA2 for ' + \
     'the Time-Dependent Vehicle Routing Problem with Time Windows ' + \
     '(TDVRPTW), as proposed by: \n\nNanda Kumar, Suresh & Panneerselvam, ' + \
     'Ramasamy. (2017). Development of an Efficient Genetic Algorithm for' + \
     'the Time Dependent Vehicle Routing Problem with Time Windows. ' + \
     'American Journal of Operations Research. 07. 1-25. ' + \
     '10.4236/ajor.2017.71001.\n\nUsage instructions can be found here: ' + \
     'https://github.com/yma17/tdvrptw-snrpga2'

setup(
    name="tdvrptw_snrpga2", 
    version=VERSION,
    author="Yingchen Ma",
    author_email="ericma@comcast.net",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_package_data=True,
    install_requires=['numpy', 'scipy', 'tqdm'],
    keywords=['python', 'tdvrptw', 'vrp', 'tdvrp', 'vrptw'],
    classifiers= [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


"""
Contains class definition of a TDVRPTW instance.
"""

import os
import time
import numpy as np
join = os.path.join

from .compute_inputs import *
from .genetic_op import eval_fitness
from .genetic_alg import snrpga2


class TDVRPTWInstance(object):
    """Instance of a TDVRPTW problem."""

    def __init__(self, inst_name='Instance 1'):
        """Initialize algorithm inputs and hyperparameters."""

        #
        # Hyperparameters
        #
        self.params = {}

        # Tournament selection probability
        self.params["ts_prob"] = 0.9
        # Crossover probability
        self.params["x_prob"] = 0.7
        # Mutation probability
        self.params["m_prob"] = 0.3
        # Weight of time in objective function
        self.params["w_t"] = 1.0
        # Maximum number of generations (e.g. number of alg iterations)
        self.params["mng"] = 1000
        # Number of individuals in initial population
        self.params["pop_size"] = 100
        # Method of initialization of first generation
        self.params["init_method"] = 'random_sample'
        #self.params["init_method"] = 'random_seq' 
        # Objective function
        self.params["obj_func"] = 'dt'  # weighted sum of distance and time
        #self.params["obj_func"] = 'd'  # distance only
        #self.params["obj_func"] = 't'  # time only
        #self.params["obj_func"] = 'v'  # number of vehicles
        # Granularity of keys of time to index cache
        self.params["cache_gran"] = 1.0

        #
        # Algorithm Inputs
        #

        self.name = inst_name  # instance name

        self.num_vehicles = 0  # (maximum) number of vehicles (optional)
        self.C = None  # maximum truck capacity

        self.location_ids = []  # list of location id names
        self.x, self.y = [], []  # coordinates for each location
        self.demand = []  # demand for each location
        self.ready_time = []  # start time for each location's time window
        self.due_time = []  # end time for each location's time window
        self.service_time = []  # necessary service time for each location

        self.D_m = None  # time-dependent (3D) distance matrix
        self.T_m = None  # time-dependent (3D) time matrix
        self.t2i = None  # map time to axis 0 index of self.D_m and self.T_m
        self.i2t = None  # list of matrix time values

        # Speed and category matrices (optional)
        # Ichoua et al. 2003, Balseiro et al. 2010 only
        self.speed_mat = None
        self.cat_mat = None


    def get_param(self, param_name):
        """Getter function for a hyperparameter."""
        assert param_name in self.params.keys()
        return self.params[param_name]


    def get_params(self):
        """Getter function for all hyperparameters."""
        return self.params


    def set_solomon(self, inst_s):
        """Set parameters from Solomon (1987) instance."""

        self.name = inst_s["name"]
        self.num_vehicles = inst_s["num_vehicles"]
        self.C = inst_s["capacity"]
        self.location_ids = inst_s["customer_ids"]
        self.x, self.y = inst_s["x"], inst_s["y"]
        self.demand = inst_s["demand"]
        self.ready_time = inst_s["ready_time"]
        self.due_time = inst_s["due_time"]
        self.service_time = inst_s["service_time"]


    def set_ichoua(self, speed_mat, cat_mat):
        """Set matrices from Ichoua et al. (2003) data."""
        self.speed_mat = speed_mat
        self.cat_mat = cat_mat


    def set_balseiro(self, cat_mat):
        """Set matrix from Balseiro et al. (2010)."""
        self.cat_mat = cat_mat

    
    def set_param(self, param_name, value):
        """Setter function for a hyperparameter."""

        assert param_name in self.params.keys()
        if param_name == "obj_func":
            assert value in ['v', 't', 'd', 'dt']
        elif param_name == "init_method":
            assert value in ['random_sample', 'random_seq']
        
        self.params[param_name] = value
    

    def set_matrices(self, D_m, T_m, time_list):
        """
        Set time-dependent distance and time matrices,
        along with mapping from time value to axis 0 index.
        Parameters:
        D_m, T_m: np.ndarray with 3 dimensions
        time_list: list of time values corresponding to each index
                    of axis 0 in D_m and T_m
        """
        assert D_m.ndim == 3 and T_m.ndim == 3
        assert D_m.shape == T_m.shape
        assert D_m.shape[0] == len(time_list)
        assert D_m.shape[1] == len(self.x)

        self.D_m = D_m
        self.T_m = T_m
        self.i2t = time_list
        self.t2i = {self.i2t[i]:i for i in range(len(self.i2t))}
    

    def set_location_info(self,
                          x,
                          y,
                          customer_ids,
                          demand,
                          ready_time,
                          due_time,
                          service_time):
        """
        Set coordinate information, time window, service time, demand,
        and ids for locations.
        """
        assert len(x) == len(y)
        assert len(y) == len(customer_ids)
        assert len(customer_ids) == len(demand)
        assert len(demand) == len(ready_time)
        assert len(ready_time) == len(due_time)
        assert len(due_time) == len(service_time)

        self.x, self.y = x, y
        self.location_ids = customer_ids
        self.demand = np.array(demand)
        self.ready_time = np.array(ready_time)
        self.due_time = np.array(due_time)
        self.service_time = np.array(service_time)


    def run(self, scen=None):
        """
        Run data preprocessing and algorithm.
        Parameter: scen: int (scenario number; required if speed matr used)
        Returns: sol_instr (output of reformat())
        """

        print(self.get_params())

        # Compute matrices, time indices if not explicitly specified
        if self.t2i is None:
            w_s = compute_window_size(self.ready_time[0],
                                      self.due_time[0],
                                      len(self.ready_time))
            self.t2i = compute_t2i(self.ready_time[0], self.due_time[0], w_s)
            D_r = compute_dist_matrix(self.x, self.y)
            self.D_m = np.repeat(D_r[np.newaxis, :, :], len(self.t2i), axis=0)
            self.i2t = compute_i2t(self.t2i)
            T_r = compute_raw_time_matrix(D_r,
                                          scen,
                                          self.cat_mat,
                                          self.speed_mat,
                                          self.i2t)
            self.T_m = compute_time_matrix(T_r, self.ready_time, self.i2t)

        # Call algorithm
        begin_time = time.time()
        res, score = snrpga2(self.D_m, self.T_m, self.service_time,
                                self.demand, self.due_time, self.ready_time[0],
                                self.t2i, self.i2t, self.C, self.params)
        end_time = time.time()

        d, t = [eval_fitness(f,
                             g_start=self.ready_time[0],
                             routes=res[0],
                             departure_times=res[3],
                             dist_matrix=self.D_m,
                             times=self.i2t,
                             t2i=self.t2i,
                             cache={},
                             granularity=self.params["cache_gran"],
                             depot_arrivals=res[4],
                             w_t=self.params["w_t"])
                             for f in ('d', 't')]

        sol_instr = self.reformat(res)

        print("--- RESULTS FOR {} ---".format(self.name))
        print("Score: {}".format(score))
        print("Distance to travel: {}".format(d))
        print("Total travel time: {}".format(t))
        print("Runtime: {} seconds".format(end_time - begin_time))
        print("Number of trucks used: {}/{}".format(len(sol_instr["per_truck"]),
                self.num_vehicles))
        print("Routes:", res[0])
        print()

        return sol_instr


    def reformat(self, sol_res):
        """
        Reformat algorithm output into sequential, more human-readable format.
        Parameters:
        - sol_res: tuple (output of tdvrptw_snrpga2.genetic_alg.snrpga2())
        Returns:
        - sol_instr: dict (instructions)
        """

        sol_instr = {}

        # Information for each truck
        per_truck = []
        for route_idx in range(len(sol_res[0])):
            truck_info = {}

            # Instructions for delivery to each customer
            truck_per_customer = []
            for loc_idx in range(len(sol_res[0][route_idx])):
                loc_dict = {}
                loc_dict['loc_id'] = self.location_ids[sol_res[0][route_idx][loc_idx]]
                loc_dict['deliv_amount'] = sol_res[1][route_idx][loc_idx]
                loc_dict['arrival_t'] = sol_res[2][route_idx][loc_idx]
                loc_dict['depart_t'] = sol_res[3][route_idx][loc_idx]

                truck_per_customer += [loc_dict]
            truck_info["per_customer"] = truck_per_customer

            # Instructions for travel back to depot
            loc_dict = {}
            loc_dict['loc_id'] = self.location_ids[0]
            loc_dict['arrival_t'] = sol_res[4][route_idx]
            truck_info["back_to_depot"] = loc_dict

            # Specify statistics for truck across all locations
            truck_overall = {}
            truck_overall["num_customers"] = len(sol_res[0][route_idx])
            truck_overall["total_deliv_amount"] = sum(sol_res[1][route_idx])
            all_truck_times = [truck_info["back_to_depot"]["arrival_t"]] + \
                [x["arrival_t"] for x in truck_info["per_customer"]] + \
                [x["depart_t"] for x in truck_info["per_customer"]]
            truck_overall["earliest_time"] = min(all_truck_times)
            truck_overall["latest_time"] = max(all_truck_times)
            truck_info["overall"] = truck_overall

            per_truck += [truck_info]
        sol_instr["per_truck"] = per_truck

        # Statistics across all trucks
        sol_instr["overall"] = {}
        sol_instr["overall"]["num_trucks"] = len(sol_res[0])
        sol_instr["overall"]["num_customers"] = sum([x["overall"]["num_customers"] for x in per_truck])
        sol_instr["overall"]["total_deliv_amount"] = sum([x["overall"]["total_deliv_amount"] for x in per_truck])
        sol_instr["overall"]["earliest_time"] = min([x["overall"]["earliest_time"] for x in per_truck])
        sol_instr["overall"]["latest_time"] = max([x["overall"]["latest_time"] for x in per_truck])

        return sol_instr


"""
Contains helper functions for SNRPGA2 genetic algorithm operations.
"""

import math
import numpy as np
from random import random, randint, sample

from .utils import take_closest


def subpop_selection(scores, n=0.3, mode='prop'):
    """
    "SUBPOPULATION SELECTION" MODULE OF GENETIC ALGORITHM.
    Sorts population by fitness score in ascending order.
    Then, takes a subset of the sorted scores to form a subpopulation.

    Parameters:
    - scores: list of float (already-computed fitness scores)
    - n: float or int (proportion or number to keep)
    - mode: str ('prop' or 'top')

    Returns:
    - subpop_ind: list of int (indices of subpopulation chromosomes)
    """
    assert mode == 'prop' or mode == 'top'
    if mode == 'prop':
        assert n > 0 and n <= 1
    else:
        assert isinstance(n, int) and n >= 1

    # Sort by fitness score
    ind = list(range(len(scores)))
    _, sorted_ind = zip(*sorted(zip(scores, ind)))

    # Retrieve top proportion or top number
    if mode == 'prop':
        subpop_ind = list(sorted_ind)[:math.ceil(len(scores) * n)]
    else:
        subpop_ind = list(sorted_ind)[:n]

    return subpop_ind


def tournament_selection(S_ind, n, k, p=0.9):
    """
    "TOURNAMENT SELECTION" MODULE OF GENETIC ALGORITHM.
    Select pairs of two for crossover.

    Parameters:
    - S_ind: list of int (indices of subpopulation)
        - output of subpop_selection()
    - n: int (number of tournaments to run)
    - k: int (number of individuals in each tournament)
    - p: float (probability of selecting indiv with best fitness)

    Returns:
    - sel_pairs: list of np.ndarray (index pairs)
    """

    prob = [p * ((1 - p) ** i) for i in range(k)]
    prob[-1] = 1.0 - sum(prob[:-1])

    # sel_pairs = []
    # x_list = np.random.choice(range(len(S_ind)), len(S_ind), replace=False)
    # for i in range(0, len(x_list) - 1, 2):
    #     sel_pairs += [np.array([x_list[i], x_list[i + 1]])]

    sel_pairs = []
    sel_set = set()
    for _ in range(n):
        # Get indices of contestants
        x_list = sample(range(len(S_ind)), k)
        x_list.sort()
        C_ind = [S_ind[x] for x in x_list]

        # Select pair of two contestants
        if p < 1.0:
            pair = np.random.choice(C_ind, 2, replace=False, p=prob)
        else:
            pair = C_ind[:2]
        sel_pairs += [pair]
        sel_set.add(pair[0])
        sel_set.add(pair[1])
    
    # Ensure every chromosome in subpopulation is included
    #   in at least one pair
    ind_left = list(set(S_ind) - sel_set)
    for i in range(0, len(ind_left) - 1, 2):
        sel_pairs += [np.array([ind_left[i], ind_left[i + 1]])]
    if len(ind_left) % 2 == 0:
        sel_pairs += [np.array([S_ind[0], ind_left[-1]])]

    # # Ensure every chromosome in subpopulation is included
    # #   in at least one pair
    # ind_left = list(set(S_ind) - sel_set)
    # S_ind_k = S_ind[:k]
    # for i in range(len(ind_left)):
    #     x = np.random.choice(S_ind_k, 1, p=prob)
    #     sel_pairs += [np.array([x[0], ind_left[i]])]

    return sel_pairs


def crossover(c1, c2, x_prob=0.7):
    """
    "CROSSOVER" MODULE OF GENETIC ALGORITHM.
    Performs random sequence iteration-based crossover (RSIX).
    Parameters:
    - c1, c2: np.ndarray (representing parent chromosomes)
    - prob: float (crossover probability)
    Returns:
    - d1, d2: np.ndarray (representing offspring chromosomes)
    - bool (True if offspring are different than parents)
    """

    p = random()
    if p >= x_prob:
        return c1, c2, False

    # Generate two crossover points of length three
    c1_i, c2_i = randint(0, c1.shape[0] - 3), randint(0, c2.shape[0] - 3)
    c1_pt, c2_pt = c1[c1_i:c1_i + 3].tolist(), c2[c2_i:c2_i + 3].tolist()
    c, c_i, c_pt = [c1, c2], [c1_i, c2_i], [c1_pt, c2_pt]

    # Swapping operation, wrt problem constraints
    d = []
    for x in range(2):
        d_next = []
        for i in range(len(c[x])):
            if i == c_i[x]:
                d_next += c_pt[(x + 1) % 2]
            if c[x][i] not in c_pt[(x + 1) % 2]:
                d_next += [c[x][i]]
        d += [d_next]
    
    d1, d2 = np.array(d[0]), np.array(d[1])
    assert c1.shape[0] == d1.shape[0] and c2.shape[0] == d2.shape[0]
    return d1, d2, True


def mutation(d1, d2, m_prob=0.3):
    """
    "MUTATION" MODULE OF GENETIC ALGORITHM.
    Performs mutation on a chromosome/offspring.
    Parameters:
    - d1, d2: np.ndarray (offspring)
    - m_prob: float (mutation probability)
    Returns:
    - d1, d2: np.ndarray (mutated offspring)
    - bool (True if mutation occurred)
    """
    
    p = random()
    if p >= m_prob:
        return d1, d2, False

    # Select two genes from each chromosome to swap
    d1_i = np.random.choice(d1.shape[0], 2, replace=False)
    d2_i = np.random.choice(d2.shape[0], 2, replace=False)
    d1[d1_i[0]], d1[d1_i[1]] = d1[d1_i[1]], d1[d1_i[0]]
    d2[d2_i[0]], d2[d2_i[1]] = d2[d2_i[1]], d2[d2_i[0]]

    return d1, d2, True


def assign_routes(ch, C, D_t, D, E, T, g_start,
                    times, t2i, cache={}, granularity=1.0):
    """
    Assign a single-route chromosome into multiple sub-chromosomes
        of smaller routes consisting of a set of customers.
    These sub-chromosomes are used for fitness evaluation,
        and may represent a solution to the overall problem.
    Problem constraints are utilized here.

    Parameters:
    - ch: 1D np.ndarray (chromosome)
    - C: float (truck capacity; assume it is the same for each truck)
    - D_t: 1D np.ndarray (delivery times for each location)
    - D: 1D np.ndarray (demand for each location)
    - E: 1D np.ndarray (end time for each location)
    - T: 3D np.ndarray (time-dependent travel time matrix)
        - axes: time, source, destination (respectively)
    - g_start: float (global start time, aka departure from depot)
    - times: list (times of indices of axis 0 of T)
    - t2i: dict (map time to index of T on axis 0)
    - cache: dict
    - granularity: float

    Returns:
    - a tuple containing:
        - routes: list of lists (list of routes; each val = one truck)
        - amounts: list of 1ists (list of delivery amounts)
        - arrival_times: list of lists (arrival times to each customer)
        - departure_times: list of lists (departure times from each customer)
        - depot_arrivals: list (final arrival time at depot for each truck)
    """

    # Output variables
    routes = []
    arrival_times = []
    depot_arrivals = []
    amounts = []
    departure_times = []

    # Temporary variables
    curr_route = [0]  # locations visited so far
    curr_arrival = []  # arrival times so far
    curr_amounts = []  # amounts delivered to each location
    curr_departure = [g_start]  # departure times so far
    curr_deliv = 0.0  # total amount delivered so far

    # Assign customers iteratively, adding more trucks if needed
    cust_idx = 0
    while cust_idx < len(ch):
        next_cust = ch[cust_idx]

        # Retrieve travel time from current to next location
        t_next = take_closest(times,
                              curr_departure[-1],
                              cache=cache,
                              granularity=granularity)
        i_next = t2i[t_next]
        travel_time_next = T[i_next, curr_route[-1], next_cust]

        next_arrival = curr_departure[-1] + travel_time_next

        # Retrieve travel time from next location to depot
        next_departure = next_arrival + D_t[next_cust]
        t_depot = take_closest(times,
                               next_departure,
                               cache=cache,
                               granularity=granularity)
        i_depot = t2i[t_depot]
        travel_time_depot = T[i_depot, next_cust, 0]

        # Check customer end time, depot end time, and capacity constraints
        over_time_cust = (next_arrival > E[next_cust])
        over_time_depot = (next_departure + travel_time_depot > E[0])
        over_capacity = (curr_deliv + D[next_cust] > C)

        if not over_time_cust and not over_time_depot and not over_capacity:
            # Add to existing route information
            curr_route += [next_cust]
            curr_arrival += [curr_departure[-1] + travel_time_next]
            curr_amounts += [D[next_cust]]
            curr_departure += [next_departure]
            curr_deliv += D[next_cust]

            cust_idx += 1

            if cust_idx < len(ch):
                continue

        # Conclude current route; drive truck back to depot
        routes += [curr_route[1:]]
        arrival_times += [curr_arrival]
        amounts += [curr_amounts]
        departure_times += [curr_departure[1:]]
        depot_arrivals += [curr_departure[-1] + T[i_next, curr_route[-1], 0]]

        # Start next route; reset temporary variables
        curr_route, curr_arrival, curr_deliv = [0], [], 0.0
        curr_amounts, curr_departure = [], [g_start]

    return (routes, amounts, arrival_times, departure_times, depot_arrivals)


def eval_fitness(obj_func, g_start, routes=None, departure_times=None,
                    dist_matrix=None, times=None, t2i=None, cache=None,
                    granularity=None, depot_arrivals=None, w_t=1):
    """
    "FITNESS EVALUATION" MODULE OF GENETIC ALGORITHM.
    Evaluates fitness function with respect to distance/time/vehicles.

    Parameters:
    - obj_func: str
    - routes: list of lists
    - departure_times: list of lists
    - dist_matrix: 3D np.ndarray
    - times: list
    - t2i: dict
    - cache: dict
    - granularity: float
    - depot_arrivals: list
    - g_start: int
    - w_t: float

    Returns:
    - score: float (representing fitness score)
    """
    assert obj_func in ['v', 'd', 't', 'dt']
    
    if obj_func == 'v':  # number of vehicles
        return len(routes)

    assert g_start is not None

    score = 0.0

    if 'd' in obj_func:  # distance
        assert routes is not None and dist_matrix is not None
        assert departure_times is not None
        assert times is not None and t2i is not None
        assert cache is not None and granularity is not None

        for x, route in enumerate(routes):
            # Depot to first customer
            t = take_closest(times,
                             g_start,
                             cache=cache,
                             granularity=granularity)
            score += dist_matrix[t2i[t], 0, route[0]]

            # Customer to customer
            for y in range(len(route) - 1):
                t = take_closest(times,
                                 departure_times[x][y],
                                 cache=cache,
                                 granularity=granularity)
                score += dist_matrix[t2i[t], route[y], route[y + 1]]
            
            # Last customer (back) to depot
            t = take_closest(times,
                             departure_times[x][-1],
                             cache=cache,
                             granularity=granularity)
            score += dist_matrix[t2i[t], route[-1], 0]

    if 't' in obj_func:  # time
        assert depot_arrivals is not None

        time_score = sum(depot_arrivals) - g_start * len(depot_arrivals)
        
        mult = w_t if 'd' in obj_func else 1.0
        score += (mult * time_score)

    return score


def replace_in_pop(L, L_scores, L_res, i, o, o_score, o_res):
    """
    "REPLACE" MODULE OF GENETIC ALGORITHM.
    Replace parent chromosome with offspring if offspring has
        better fitness score than parent.

    Parameters:
    - L: list of np.ndarray (population)
    - L_scores: list of float (population scores)
    - L_res: list of tuple (population detailed info)
    - i: float (parent chromosome index)
    - o: np.ndarray (offspring chromosome)
    - o_score: float (offspring chromosome score)
    - o_res: tuple (offspring detailed info)

    Returns:
    - None
    """

    if o_score < L_scores[i]:
        L[i] = o
        L_scores[i] = o_score
        L_res[i] = o_res
